# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""The base class for the settings of U-MDO formulations using a DOE."""

from __future__ import annotations

from typing import TYPE_CHECKING

from gemseo.algos.doe.base_doe_settings import BaseDOESettings  # noqa: TC002
from gemseo.algos.doe.openturns.settings.ot_opt_lhs import OT_OPT_LHS_Settings
from gemseo.utils.seeder import SEED
from pydantic import Field
from pydantic import PositiveInt
from pydantic import model_validator

from gemseo_umdo.formulations.base_umdo_formulation_settings import (
    BaseUMDOFormulationSettings,
)

if TYPE_CHECKING:
    from typing_extensions import Self


class BaseSamplingSettings(BaseUMDOFormulationSettings):
    """The base class for the settings of U-MDO formulations using a DOE."""

    n_samples: PositiveInt | None = Field(
        default=None,
        description="""The number of samples to be generated by the DOE algorithm.

If `None`,
this field is ignored.

If not `None`,
the field `doe_algo_settings.n_samples` is ignored.
""",
    )

    doe_algo_settings: BaseDOESettings = Field(
        default=OT_OPT_LHS_Settings(n_samples=10, seed=SEED),
        description="The DOE settings.",
    )

    @model_validator(mode="after")
    def __validate_seed(self) -> Self:
        """Validate the seed."""
        doe_algo_settings = self.doe_algo_settings
        if self.n_samples is not None:
            doe_algo_settings.n_samples = self.n_samples

        if "seed" in doe_algo_settings.model_fields and doe_algo_settings.seed is None:
            doe_algo_settings.seed = SEED

        return self
